import re
import socket
import binascii
import asyncio
import logging
import typing
from collections import OrderedDict
from asyncio.futures import Future
from asyncio.transports import DatagramTransport
from aioupnp.fault import UPnPError
from aioupnp.serialization.ssdp import SSDPDatagram
from aioupnp.constants import SSDP_IP_ADDRESS, SSDP_PORT
from aioupnp.protocols.multicast import MulticastProtocol
from aioupnp.protocols.m_search_patterns import packet_generator

ADDRESS_REGEX = re.compile("^http:\/\/(\d+\.\d+\.\d+\.\d+)\:(\d*)(\/[\w|\/|\:|\-|\.]*)$")

log = logging.getLogger(__name__)


class SSDPProtocol(MulticastProtocol):
    def __init__(self, multicast_address: str, lan_address: str, ignored: typing.Set[str] = None,
                 unicast: bool = False) -> None:
        super().__init__(multicast_address, lan_address)
        self._unicast = unicast
        self._ignored: typing.Set[str] = ignored or set()  # ignored locations
        self._pending_searches: typing.List[typing.Tuple[str, str, Future, asyncio.Handle]] = []
        self.notifications: typing.List = []

    def disconnect(self):
        if self.transport:
            self.transport.close()
        while self._pending_searches:
            pending = self._pending_searches.pop()[2]
            if not pending.cancelled() and not pending.done():
                pending.cancel()

    def _callback_m_search_ok(self, address: str, packet: SSDPDatagram) -> None:
        if packet.location in self._ignored:
            return
        tmp: typing.List = []
        set_futures: typing.List = []
        while self._pending_searches:
            t: tuple = self._pending_searches.pop()
            a, s = t[0], t[1]
            if (address == a) and (s in [packet.st, "upnp:rootdevice"]):
                f: Future = t[2]
                # h: asyncio.Handle = t[3]
                # h.cancel()
                if f not in set_futures:
                    set_futures.append(f)
                    if not f.done():
                        f.set_result(packet)
            elif t[2] not in set_futures:
                tmp.append(t)
        while tmp:
            self._pending_searches.append(tmp.pop())

    def send_many_m_searches(self, address: str, packets: typing.List[SSDPDatagram]):
        dest = address if self._unicast else SSDP_IP_ADDRESS
        for packet in packets:
            log.debug("send m search to %s: %s", dest, packet.st)
            self.transport.sendto(packet.encode().encode(), (dest, SSDP_PORT))

    async def m_search(self, address: str, timeout: float, datagrams: typing.List[OrderedDict]) -> SSDPDatagram:
        fut: Future = Future()
        packets: typing.List[SSDPDatagram] = []
        for datagram in datagrams:
            packet = SSDPDatagram(SSDPDatagram._M_SEARCH, datagram)
            assert packet.st is not None
            # h = asyncio.get_running_loop().call_later(timeout, fut.cancel)
            self._pending_searches.append((address, packet.st, fut))
            packets.append(packet)
        self.send_many_m_searches(address, packets),
        return await fut

    def datagram_received(self, data, addr) -> None:
        if addr[0] == self.bind_address:
            return
        try:
            packet = SSDPDatagram.decode(data)
            log.debug("decoded packet from %s:%i: %s", addr[0], addr[1], packet)
        except UPnPError as err:
            log.error("failed to decode SSDP packet from %s:%i (%s): %s", addr[0], addr[1], err,
                      binascii.hexlify(data))
            return

        if packet._packet_type == packet._OK:
            self._callback_m_search_ok(addr[0], packet)
            return
        # elif packet._packet_type == packet._NOTIFY:
        #     log.debug("%s:%i sent us a notification: %s", packet)
        #     if packet.nt == SSDP_ROOT_DEVICE:
        #         address, port, path = ADDRESS_REGEX.findall(packet.location)[0]
        #         key = None
        #         for (addr, service) in self.discover_callbacks:
        #             if addr == address:
        #                 key = (addr, service)
        #                 break
        #         if key:
        #             log.debug("got a notification with the requested m-search info")
        #             notify_fut: Future = self.discover_callbacks.pop(key)
        #             notify_fut.set_result(SSDPDatagram(
        #                 SSDPDatagram._OK, cache_control='', location=packet.location, server=packet.server,
        #                 st=UPNP_ORG_IGD, usn=packet.usn
        #             ))
        #         self.notifications.append(packet.as_dict())
        #         return


async def listen_ssdp(lan_address: str, gateway_address: str, ssdp_socket: socket.socket = None,
                      ignored: typing.Set[str] = None, unicast: bool = False) -> typing.Tuple[DatagramTransport,
                                                                                              SSDPProtocol, str, str]:
    loop = asyncio.get_running_loop()
    try:
        sock = ssdp_socket or SSDPProtocol.create_multicast_socket(lan_address)
        listen_result: typing.Tuple = await loop.create_datagram_endpoint(
            lambda: SSDPProtocol(SSDP_IP_ADDRESS, lan_address, ignored, unicast), sock=sock
        )
        transport: DatagramTransport = listen_result[0]
        protocol: SSDPProtocol = listen_result[1]
    except Exception as err:
        raise UPnPError(err)
    try:
        protocol.join_group(protocol.multicast_address, protocol.bind_address)
        protocol.set_ttl(1)
    except Exception as err:
        protocol.disconnect()
        raise UPnPError(err)

    return transport, protocol, gateway_address, lan_address


async def m_search(lan_address: str, gateway_address: str, datagram_args: OrderedDict, timeout: int = 1,
                   ssdp_socket: socket.socket = None, ignored: typing.Set[str] = None,
                   unicast: bool = False) -> SSDPDatagram:
    transport, protocol, gateway_address, lan_address = await listen_ssdp(
        lan_address, gateway_address, ssdp_socket, ignored, unicast
    )
    try:
        return await protocol.m_search(address=gateway_address, timeout=timeout, datagrams=[datagram_args])
    except (asyncio.TimeoutError, asyncio.CancelledError):
        raise UPnPError("M-SEARCH for {}:{} timed out".format(gateway_address, SSDP_PORT))
    finally:
        protocol.disconnect()


async def _fuzzy_m_search(lan_address: str, gateway_address: str, timeout: int = 30,
                          ssdp_socket: socket.socket = None,
                          ignored: typing.Set[str] = None, unicast: bool = False) -> typing.List[OrderedDict]:
    transport, protocol, gateway_address, lan_address = await listen_ssdp(
        lan_address, gateway_address, ssdp_socket, ignored, unicast
    )
    packet_args = list(packet_generator())
    batch_size = 2
    batch_timeout = float(timeout) / float(len(packet_args))
    while packet_args:
        args = packet_args[:batch_size]
        packet_args = packet_args[batch_size:]
        log.debug("sending batch of %i M-SEARCH attempts", batch_size)
        try:
            await asyncio.wait_for(protocol.m_search(gateway_address, batch_timeout, args), timeout)
            protocol.disconnect()
            return args
        except asyncio.TimeoutError:
            continue
    protocol.disconnect()
    raise UPnPError("M-SEARCH for {}:{} timed out".format(gateway_address, SSDP_PORT))


async def fuzzy_m_search(lan_address: str, gateway_address: str, timeout: int = 30,
                         ssdp_socket: socket.socket = None,
                         ignored: typing.Set[str] = None, unicast: bool = False) -> typing.Tuple[OrderedDict,
                                                                                                 SSDPDatagram]:
    # we don't know which packet the gateway replies to, so send small batches at a time

    args_to_try = await _fuzzy_m_search(lan_address, gateway_address, timeout, ssdp_socket, ignored, unicast)
    # check the args in the batch that got a reply one at a time to see which one worked
    for args in args_to_try:
        try:
            packet = await m_search(lan_address, gateway_address, args, 3, ignored=ignored, unicast=unicast)
            return args, packet
        except UPnPError:
            continue
    raise UPnPError("failed to discover gateway")



import re
import socket
from collections import defaultdict
from typing import Tuple, Dict
from xml.etree import ElementTree
import netifaces


BASE_ADDRESS_REGEX = re.compile("^(http:\/\/\d*\.\d*\.\d*\.\d*:\d*)\/.*$".encode())
BASE_PORT_REGEX = re.compile("^http:\/\/\d*\.\d*\.\d*\.\d*:(\d*)\/.*$".encode())


def etree_to_dict(t: ElementTree.Element) -> Dict:
    d: dict = {}
    if t.attrib:
        d[t.tag] = {}
    children = list(t)
    if children:
        dd: dict = defaultdict(list)
        for dc in map(etree_to_dict, children):
            for k, v in dc.items():
                dd[k].append(v)
        d[t.tag] = {k: v[0] if len(v) == 1 else v for k, v in dd.items()}
    if t.attrib:
        d[t.tag].update(('@' + k, v) for k, v in t.attrib.items())
    if t.text:
        text = t.text.strip()
        if children or t.attrib:
            if text:
                d[t.tag]['#text'] = text
        else:
            d[t.tag] = text
    return d


def flatten_keys(d, strip):
    if not isinstance(d, (list, dict)):
        return d
    if isinstance(d, list):
        return [flatten_keys(i, strip) for i in d]
    t = {}
    for k, v in d.items():
        if strip in k and strip != k:
            t[k.split(strip)[1]] = flatten_keys(v, strip)
        else:
            t[k] = flatten_keys(v, strip)
    return t


def get_dict_val_case_insensitive(d, k):
    match = list(filter(lambda x: x.lower() == k.lower(), d.keys()))
    if not match:
        return
    if len(match) > 1:
        raise KeyError("overlapping keys")
    return d[match[0]]

# import struct
# import fcntl
# def get_ip_address(ifname):
#     SIOCGIFADDR = 0x8915
#     s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
#     return socket.inet_ntoa(fcntl.ioctl(
#         s.fileno(),
#         SIOCGIFADDR,
#         struct.pack(b'256s', ifname[:15].encode())
#     )[20:24])


def get_interfaces():
    r = {
        interface_name: (router_address, netifaces.ifaddresses(interface_name)[netifaces.AF_INET][0]['addr'])
        for router_address, interface_name, _ in netifaces.gateways()[socket.AF_INET]
    }
    for interface_name in netifaces.interfaces():
        if interface_name in ['lo', 'localhost'] or interface_name in r:
            continue
        addresses = netifaces.ifaddresses(interface_name)
        if netifaces.AF_INET in addresses:
            address = addresses[netifaces.AF_INET][0]['addr']
            gateway_guess = ".".join(address.split(".")[:-1] + ["1"])
            r[interface_name] = (gateway_guess, address)
    r['default'] = r[netifaces.gateways()['default'][netifaces.AF_INET][1]]
    return r


def get_gateway_and_lan_addresses(interface_name: str) -> Tuple[str, str]:
    for iface_name, (gateway, lan) in get_interfaces().items():
        if interface_name == iface_name:
            return gateway, lan
    return '', ''



from typing import Tuple, Union

none_or_str = Union[None, str]


class SOAPCommands:
    @staticmethod
    async def AddPortMapping(NewRemoteHost: str, NewExternalPort: int, NewProtocol: str, NewInternalPort: int,
                       NewInternalClient: str, NewEnabled: int, NewPortMappingDescription: str,
                       NewLeaseDuration: str) -> None:
        """Returns None"""
        raise NotImplementedError()

    @staticmethod
    async def GetNATRSIPStatus() -> Tuple[bool, bool]:
        """Returns (NewRSIPAvailable, NewNATEnabled)"""
        raise NotImplementedError()

    @staticmethod
    async def GetGenericPortMappingEntry(NewPortMappingIndex: int) -> Tuple[str, int, str, int, str,
                                                                            bool, str, int]:
        """
        Returns (NewRemoteHost, NewExternalPort, NewProtocol, NewInternalPort, NewInternalClient, NewEnabled,
                 NewPortMappingDescription, NewLeaseDuration)
        """
        raise NotImplementedError()

    @staticmethod
    async def GetSpecificPortMappingEntry(NewRemoteHost: str, NewExternalPort: int,
                                          NewProtocol: str) -> Tuple[int, str, bool, str, int]:
        """Returns (NewInternalPort, NewInternalClient, NewEnabled, NewPortMappingDescription, NewLeaseDuration)"""
        raise NotImplementedError()

    @staticmethod
    async def SetConnectionType(NewConnectionType: str) -> None:
        """Returns None"""
        raise NotImplementedError()

    @staticmethod
    async def GetExternalIPAddress() -> str:
        """Returns (NewExternalIPAddress)"""
        raise NotImplementedError()

    @staticmethod
    async def GetConnectionTypeInfo() -> Tuple[str, str]:
        """Returns (NewConnectionType, NewPossibleConnectionTypes)"""
        raise NotImplementedError()

    @staticmethod
    async def GetStatusInfo() -> Tuple[str, str, int]:
        """Returns (NewConnectionStatus, NewLastConnectionError, NewUptime)"""
        raise NotImplementedError()

    @staticmethod
    async def ForceTermination() -> None:
        """Returns None"""
        raise NotImplementedError()

    @staticmethod
    async def DeletePortMapping(NewRemoteHost: str, NewExternalPort: int, NewProtocol: str) -> None:
        """Returns None"""
        raise NotImplementedError()

    @staticmethod
    async def RequestConnection() -> None:
        """Returns None"""
        raise NotImplementedError()

    @staticmethod
    async def GetCommonLinkProperties():
        """Returns (NewWANAccessType, NewLayer1UpstreamMaxBitRate, NewLayer1DownstreamMaxBitRate, NewPhysicalLinkStatus)"""
        raise NotImplementedError()

    @staticmethod
    async def GetTotalBytesSent():
        """Returns (NewTotalBytesSent)"""
        raise NotImplementedError()

    @staticmethod
    async def GetTotalBytesReceived():
        """Returns (NewTotalBytesReceived)"""
        raise NotImplementedError()

    @staticmethod
    async def GetTotalPacketsSent():
        """Returns (NewTotalPacketsSent)"""
        raise NotImplementedError()

    @staticmethod
    def GetTotalPacketsReceived():
        """Returns (NewTotalPacketsReceived)"""
        raise NotImplementedError()

    @staticmethod
    async def X_GetICSStatistics() -> Tuple[int, int, int, int, str, str]:
        """Returns (TotalBytesSent, TotalBytesReceived, TotalPacketsSent, TotalPacketsReceived, Layer1DownstreamMaxBitRate, Uptime)"""
        raise NotImplementedError()

    @staticmethod
    async def GetDefaultConnectionService():
        """Returns (NewDefaultConnectionService)"""
        raise NotImplementedError()

    @staticmethod
    async def SetDefaultConnectionService(NewDefaultConnectionService: str) -> None:
        """Returns (None)"""
        raise NotImplementedError()

    @staticmethod
    async def SetEnabledForInternet(NewEnabledForInternet: bool) -> None:
        raise NotImplementedError()

    @staticmethod
    async def GetEnabledForInternet() -> bool:
        raise NotImplementedError()

    @staticmethod
    async def GetMaximumActiveConnections(NewActiveConnectionIndex: int):
        raise NotImplementedError()

    @staticmethod
    async def GetActiveConnections() -> Tuple[str, str]:
        """Returns (NewActiveConnDeviceContainer, NewActiveConnectionServiceID"""
        raise NotImplementedError()



__version__ = "0.0.2a9"
__name__ = "aioupnp"
__author__ = "Jack Robison"
__maintainer__ = "Jack Robison"
__license__ = "MIT"
__email__ = "jackrobison@lbry.io"

import sys
import asyncio

if not hasattr(asyncio, "get_running_loop") and hasattr(asyncio, "_get_running_loop"):  # python 3.6
    setattr(sys.modules['asyncio'], "get_running_loop", asyncio._get_running_loop)



import logging
import socket
import typing
import re
from collections import OrderedDict
from xml.etree import ElementTree
import asyncio
from asyncio.protocols import Protocol
from aioupnp.fault import UPnPError
from aioupnp.util import get_dict_val_case_insensitive
from aioupnp.serialization.scpd import deserialize_scpd_get_response
from aioupnp.serialization.scpd import serialize_scpd_get
from aioupnp.serialization.soap import serialize_soap_post, deserialize_soap_post_response


log = logging.getLogger(__name__)


HTTP_CODE_REGEX = re.compile(b"^HTTP[\/]{0,1}1\.[1|0] (\d\d\d)(.*)$")


def parse_headers(response: bytes) -> typing.Tuple[OrderedDict, int, bytes]:
    lines = response.split(b'\r\n')
    headers = OrderedDict([
        (l.split(b':')[0], b':'.join(l.split(b':')[1:]).lstrip(b' ').rstrip(b' '))
        for l in response.split(b'\r\n')
    ])
    if len(lines) != len(headers):
        raise ValueError("duplicate headers")
    http_response = tuple(headers.keys())[0]
    response_code, message = HTTP_CODE_REGEX.findall(http_response)[0]
    del headers[http_response]
    return headers, int(response_code), message


class SCPDHTTPClientProtocol(Protocol):
    """
    This class will make HTTP GET and POST requests

    It differs from spec HTTP in that the version string can be invalid, all we care about is the xml body
    and devices respond with an invalid HTTP version line
    """

    def __init__(self, message: bytes, finished: asyncio.Future, soap_method: str=None,
                 soap_service_id: str=None) -> None:
        self.message = message
        self.response_buff = b""
        self.finished = finished
        self.soap_method = soap_method
        self.soap_service_id = soap_service_id

        self._response_code: int = 0
        self._response_msg: bytes = b""
        self._content_length: int = 0
        self._got_headers = False
        self._headers: dict = {}
        self._body = b""

    def connection_made(self, transport):
        transport.write(self.message)

    def data_received(self, data):
        self.response_buff += data
        for i, line in enumerate(self.response_buff.split(b'\r\n')):
            if not line:  # we hit the blank line between the headers and the body
                if i == (len(self.response_buff.split(b'\r\n')) - 1):
                    continue  # the body is still yet to be written
                if not self._got_headers:
                    self._headers, self._response_code, self._response_msg = parse_headers(
                        b'\r\n'.join(self.response_buff.split(b'\r\n')[:i])
                    )
                    content_length = get_dict_val_case_insensitive(self._headers, b'Content-Length')
                    if content_length is None:
                        return
                    self._content_length = int(content_length or 0)
                    self._got_headers = True
                body = b'\r\n'.join(self.response_buff.split(b'\r\n')[i+1:])
                if self._content_length == len(body):
                    self.finished.set_result((body, self._response_code, self._response_msg))
                elif self._content_length > len(body):
                    pass
                else:
                    self.finished.set_exception(
                        UPnPError(
                                "too many bytes written to response (%i vs %i expected)" % (
                                 len(body), self._content_length
                                )
                        )
                    )
                return


async def scpd_get(control_url: str, address: str, port: int) -> typing.Tuple[typing.Dict, bytes,
                                                                              typing.Optional[Exception]]:
    loop = asyncio.get_running_loop()
    finished: asyncio.Future = asyncio.Future()
    packet = serialize_scpd_get(control_url, address)
    transport, protocol = await loop.create_connection(
        lambda : SCPDHTTPClientProtocol(packet, finished),  address, port
    )
    assert isinstance(protocol, SCPDHTTPClientProtocol)
    error = None
    try:
        body, response_code, response_msg = await asyncio.wait_for(finished, 1.0)
    except asyncio.TimeoutError:
        error = UPnPError("get request timed out")
        body = b''
    finally:
        transport.close()
    if not error:
        try:
            return deserialize_scpd_get_response(body), body, None
        except ElementTree.ParseError as err:
            error = UPnPError(err)
    return {}, body, error


async def scpd_post(control_url: str, address: str, port: int, method: str, param_names: list, service_id: bytes,
                    soap_socket: socket.socket = None, **kwargs) -> typing.Tuple[typing.Dict, bytes,
                                                                                 typing.Optional[Exception]]:
    loop = asyncio.get_running_loop()
    finished: asyncio.Future = asyncio.Future()
    packet = serialize_soap_post(method, param_names, service_id, address.encode(), control_url.encode(), **kwargs)
    transport, protocol = await loop.create_connection(
        lambda : SCPDHTTPClientProtocol(
            packet, finished, soap_method=method, soap_service_id=service_id.decode(),
        ), address, port, sock=soap_socket
    )
    assert isinstance(protocol, SCPDHTTPClientProtocol)
    try:
        body, response_code, response_msg = await asyncio.wait_for(finished, 1.0)
    except asyncio.TimeoutError:
        return {}, b'', UPnPError("Timeout")
    finally:
        transport.close()
    try:
        return (
            deserialize_soap_post_response(body, method, service_id.decode()), body, None
        )
    except (ElementTree.ParseError, UPnPError) as err:
        return {}, body, UPnPError(err)



POST = "POST"
ROOT = "root"
SPEC_VERSION = "specVersion"
XML_VERSION = "<?xml version=\"1.0\"?>"
FAULT = "{http://schemas.xmlsoap.org/soap/envelope/}Fault"
ENVELOPE = "{http://schemas.xmlsoap.org/soap/envelope/}Envelope"
BODY = "{http://schemas.xmlsoap.org/soap/envelope/}Body"


CONTROL = 'urn:schemas-upnp-org:control-1-0'
SERVICE = 'urn:schemas-upnp-org:service-1-0'
DEVICE = 'urn:schemas-upnp-org:device-1-0'

WIFI_ALLIANCE_ORG_IGD = "urn:schemas-wifialliance-org:device:WFADevice:1"
UPNP_ORG_IGD = 'urn:schemas-upnp-org:device:InternetGatewayDevice:1'

WAN_SCHEMA = 'urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1'
LAYER_SCHEMA = 'urn:schemas-upnp-org:service:Layer3Forwarding:1'
IP_SCHEMA = 'urn:schemas-upnp-org:service:WANIPConnection:1'

SSDP_IP_ADDRESS = '239.255.255.250'
SSDP_PORT = 1900
SSDP_HOST = "%s:%i" % (SSDP_IP_ADDRESS, SSDP_PORT)
SSDP_DISCOVER = "ssdp:discover"
line_separator = "\r\n"



import unittest
from collections import OrderedDict
from aioupnp.serialization.ssdp import SSDPDatagram
from aioupnp.fault import UPnPError
from aioupnp.constants import UPNP_ORG_IGD, SSDP_DISCOVER


class TestMSearchDatagramSerialization(unittest.TestCase):
    packet = \
        b'M-SEARCH * HTTP/1.1\r\n' \
        b'Host: 239.255.255.250:1900\r\n' \
        b'Man: "ssdp:discover"\r\n' \
        b'ST: ssdp:all\r\n' \
        b'MX: 5\r\n' \
        b'\r\n'

    datagram_args = OrderedDict([
        ('Host', "{}:{}".format('239.255.255.250', 1900)),
        ('Man', '"ssdp:discover"'),
        ('ST', 'ssdp:all'),
        ('MX', 5),
    ])

    def test_deserialize_and_reserialize(self):
        packet1 = SSDPDatagram.decode(self.packet)
        packet2 = SSDPDatagram("M-SEARCH", self.datagram_args)
        self.assertEqual(packet2.encode(), packet1.encode())


class TestSerializationOrder(TestMSearchDatagramSerialization):
    packet = \
        b'M-SEARCH * HTTP/1.1\r\n' \
        b'Host: 239.255.255.250:1900\r\n' \
        b'ST: ssdp:all\r\n' \
        b'Man: "ssdp:discover"\r\n' \
        b'MX: 5\r\n' \
        b'\r\n'

    datagram_args = OrderedDict([
        ('Host', "{}:{}".format('239.255.255.250', 1900)),
        ('ST', 'ssdp:all'),
        ('Man', '"ssdp:discover"'),
        ('MX', 5),
    ])


class TestSerializationPreserveCase(TestMSearchDatagramSerialization):
    packet = \
        b'M-SEARCH * HTTP/1.1\r\n' \
        b'HOST: 239.255.255.250:1900\r\n' \
        b'ST: ssdp:all\r\n' \
        b'Man: "ssdp:discover"\r\n' \
        b'mx: 5\r\n' \
        b'\r\n'

    datagram_args = OrderedDict([
        ('HOST', "{}:{}".format('239.255.255.250', 1900)),
        ('ST', 'ssdp:all'),
        ('Man', '"ssdp:discover"'),
        ('mx', 5),
    ])


class TestSerializationPreserveAllLowerCase(TestMSearchDatagramSerialization):
    packet = \
        b'M-SEARCH * HTTP/1.1\r\n' \
        b'host: 239.255.255.250:1900\r\n' \
        b'st: ssdp:all\r\n' \
        b'man: "ssdp:discover"\r\n' \
        b'mx: 5\r\n' \
        b'\r\n'

    datagram_args = OrderedDict([
        ('host', "{}:{}".format('239.255.255.250', 1900)),
        ('st', 'ssdp:all'),
        ('man', '"ssdp:discover"'),
        ('mx', 5),
    ])


class TestParseMSearchRequestWithQuotes(unittest.TestCase):
    datagram = b"M-SEARCH * HTTP/1.1\r\n" \
               b"HOST: 239.255.255.250:1900\r\n" \
               b"ST: urn:schemas-upnp-org:device:InternetGatewayDevice:1\r\n" \
               b"MAN: \"ssdp:discover\"\r\n" \
               b"MX: 1\r\n" \
               b"\r\n"

    def test_parse_m_search(self):
        packet = SSDPDatagram.decode(self.datagram)
        self.assertTrue(packet._packet_type, packet._M_SEARCH)
        self.assertEqual(packet.host, '239.255.255.250:1900')
        self.assertEqual(packet.st, 'urn:schemas-upnp-org:device:InternetGatewayDevice:1')
        self.assertEqual(packet.man, '"ssdp:discover"')
        self.assertEqual(packet.mx, 1)


class TestParseMSearchRequestWithoutQuotes(unittest.TestCase):
    datagram = b'M-SEARCH * HTTP/1.1\r\n' \
               b'HOST: 239.255.255.250:1900\r\n' \
               b'ST: urn:schemas-upnp-org:device:InternetGatewayDevice:1\r\n' \
               b'MAN: ssdp:discover\r\n' \
               b'MX: 1\r\n' \
               b'\r\n'

    def test_parse_m_search(self):
        packet = SSDPDatagram.decode(self.datagram)
        self.assertTrue(packet._packet_type, packet._M_SEARCH)
        self.assertEqual(packet.host, '239.255.255.250:1900')
        self.assertEqual(packet.st, 'urn:schemas-upnp-org:device:InternetGatewayDevice:1')
        self.assertEqual(packet.man, 'ssdp:discover')
        self.assertEqual(packet.mx, 1)

    def test_serialize_m_search(self):
        packet = SSDPDatagram.decode(self.datagram)
        self.assertEqual(packet.encode().encode(), self.datagram)


class TestParseMSearchResponse(unittest.TestCase):
    datagram = "\r\n".join([
        'HTTP/1.1 200 OK',
        'CACHE_CONTROL: max-age=1800',
        'LOCATION: http://10.0.0.1:49152/InternetGatewayDevice.xml',
        'SERVER: Linux, UPnP/1.0, DIR-890L Ver 1.20',
        'ST: urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1',
        'USN: uuid:00000000-0000-0000-0000-000000000000::urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1'
    ]).encode()

    def test_parse_m_search_response(self):
        packet = SSDPDatagram.decode(self.datagram)
        self.assertTrue(packet._packet_type, packet._OK)
        self.assertEqual(packet.cache_control, 'max-age=1800')
        self.assertEqual(packet.location, 'http://10.0.0.1:49152/InternetGatewayDevice.xml')
        self.assertEqual(packet.server, 'Linux, UPnP/1.0, DIR-890L Ver 1.20')
        self.assertEqual(packet.st, 'urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1')
        self.assertEqual(packet.usn, 'uuid:00000000-0000-0000-0000-000000000000::urn:'
                                     'schemas-upnp-org:service:WANCommonInterfaceConfig:1'
)


class TestParseMSearchResponseRedSonic(TestParseMSearchResponse):
    datagram = \
        b"HTTP/1.1 200 OK\r\n" \
        b"CACHE-CONTROL: max-age=1800\r\n" \
        b"DATE: Thu, 04 Oct 2018 22:59:40 GMT\r\n" \
        b"EXT:\r\n" \
        b"LOCATION: http://10.1.10.1:49152/IGDdevicedesc_brlan0.xml\r\n" \
        b"OPT: \"http://schemas.upnp.org/upnp/1/0/\"; ns=01\r\n" \
        b"01-NLS: 00000000-0000-0000-0000-000000000000\r\n" \
        b"SERVER: Linux/3.14.28-Prod_17.2, UPnP/1.0, Portable SDK for UPnP devices/1.6.22\r\n" \
        b"X-User-Agent: redsonic\r\n" \
        b"ST: urn:schemas-upnp-org:device:InternetGatewayDevice:1\r\n" \
        b"USN: uuid:00000000-0000-0000-0000-000000000000::urn:schemas-upnp-org:device:InternetGatewayDevice:1\r\n" \
        b"\r\n"

    def test_parse_m_search_response(self):
        packet = SSDPDatagram.decode(self.datagram)
        self.assertTrue(packet._packet_type, packet._OK)
        self.assertEqual(packet.cache_control, 'max-age=1800')
        self.assertEqual(packet.location, 'http://10.1.10.1:49152/IGDdevicedesc_brlan0.xml')
        self.assertEqual(packet.server, 'Linux/3.14.28-Prod_17.2, UPnP/1.0, Portable SDK for UPnP devices/1.6.22')
        self.assertEqual(packet.st, UPNP_ORG_IGD)


class TestParseMSearchResponseDashCacheControl(TestParseMSearchResponse):
    datagram = "\r\n".join([
        'HTTP/1.1 200 OK',
        'CACHE-CONTROL: max-age=1800',
        'LOCATION: http://10.0.0.1:49152/InternetGatewayDevice.xml',
        'SERVER: Linux, UPnP/1.0, DIR-890L Ver 1.20',
        'ST: urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1',
        'USN: uuid:00000000-0000-0000-0000-000000000000::urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1'
    ]).encode()


class TestParseMSearchResponseCaseInsensitive(TestParseMSearchResponse):
    datagram = "\r\n".join([
        'HTTP/1.1 200 OK',
        'cache-CONTROL: max-age=1800',
        'LOCATION: http://10.0.0.1:49152/InternetGatewayDevice.xml',
        'Server: Linux, UPnP/1.0, DIR-890L Ver 1.20',
        'st: urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1',
        'USN: uuid:00000000-0000-0000-0000-000000000000::urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1'
    ]).encode()

    def test_get_case_insensitive(self):
        packet = SSDPDatagram.decode(self.datagram)
        self.assertEqual('max-age=1800', packet['Cache_Control'])

    def test_key_error(self):
        packet = SSDPDatagram.decode(self.datagram)
        self.assertRaises(KeyError, lambda : packet['Cache Control'])


class TestFailToParseMSearchResponseNoST(unittest.TestCase):
    datagram = "\r\n".join([
        'HTTP/1.1 200 OK',
        'CACHE_CONTROL: max-age=1800',
        'LOCATION: http://10.0.0.1:49152/InternetGatewayDevice.xml',
        'SERVER: Linux, UPnP/1.0, DIR-890L Ver 1.20',
        'USN: uuid:00000000-0000-0000-0000-000000000000::urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1'
    ]).encode()

    def test_fail_to_parse_m_search_response(self):
        self.assertRaises(UPnPError, SSDPDatagram.decode, self.datagram)


class TestFailToParseMSearchResponseNoOK(TestFailToParseMSearchResponseNoST):
    datagram = "\r\n".join([
        'cache-CONTROL: max-age=1800',
        'LOCATION: http://10.0.0.1:49152/InternetGatewayDevice.xml',
        'Server: Linux, UPnP/1.0, DIR-890L Ver 1.20',
        'st: urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1',
        'USN: uuid:00000000-0000-0000-0000-000000000000::urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1'
    ]).encode()


class TestFailToParseMSearchResponseNoLocation(TestFailToParseMSearchResponseNoST):
    datagram = "\r\n".join([
        'HTTP/1.1 200 OK',
        'cache-CONTROL: max-age=1800',
        'Server: Linux, UPnP/1.0, DIR-890L Ver 1.20',
        'st: urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1',
        'USN: uuid:00000000-0000-0000-0000-000000000000::urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1'
    ]).encode()


class TestParseNotify(unittest.TestCase):
    datagram = \
        b'NOTIFY * HTTP/1.1 \r\n' \
        b'Host: 239.255.255.250:1900\r\n' \
        b'Cache-Control: max-age=180\r\n' \
        b'Location: http://192.168.1.1:5431/dyndev/uuid:000c-29ea-247500c00068\r\n' \
        b'NT: upnp:rootdevice\r\n' \
        b'NTS: ssdp:alive\r\n' \
        b'SERVER: LINUX/2.4 UPnP/1.0 BRCM400/1.0\r\n' \
        b'USN: uuid:000c-29ea-247500c00068::upnp:rootdevice\r\n' \
        b'\r\n'

    def test_parse_notify(self):
        packet = SSDPDatagram.decode(self.datagram)
        self.assertTrue(packet._packet_type, packet._NOTIFY)
        self.assertEqual(packet.host, '239.255.255.250:1900')
        self.assertEqual(packet.cache_control, 'max-age=180')  # this is an optional field
        self.assertEqual(packet.location, 'http://192.168.1.1:5431/dyndev/uuid:000c-29ea-247500c00068')
        self.assertEqual(packet.nt, 'upnp:rootdevice')
        self.assertEqual(packet.nts, 'ssdp:alive')
        self.assertEqual(packet.server, 'LINUX/2.4 UPnP/1.0 BRCM400/1.0')
        self.assertEqual(packet.usn, 'uuid:000c-29ea-247500c00068::upnp:rootdevice')



import re
from typing import Dict
from xml.etree import ElementTree
from aioupnp.constants import XML_VERSION, DEVICE, ROOT
from aioupnp.util import etree_to_dict, flatten_keys


CONTENT_PATTERN = re.compile(
    "(\<\?xml version=\"1\.0\"\?\>(\s*.)*|\>)".encode()
)

XML_ROOT_SANITY_PATTERN = re.compile(
    "(?i)(\{|(urn:schemas-[\w|\d]*-(com|org|net))[:|-](device|service)[:|-]([\w|\d|\:|\-|\_]*)|\}([\w|\d|\:|\-|\_]*))"
)


def serialize_scpd_get(path: str, address: str) -> bytes:
    if "http://" in address:
        host = address.split("http://")[1]
    else:
        host = address
    if ":" in host:
        host = host.split(":")[0]
    if not path.startswith("/"):
        path = "/" + path
    return (
            (
                'GET %s HTTP/1.1\r\n'
                'Accept-Encoding: gzip\r\n'
                'Host: %s\r\n'
                'Connection: Close\r\n'
                '\r\n'
            ) % (path, host)
    ).encode()


def deserialize_scpd_get_response(content: bytes) -> Dict:
    if XML_VERSION.encode() in content:
        parsed = CONTENT_PATTERN.findall(content)
        content = b'' if not parsed else parsed[0][0]
        xml_dict = etree_to_dict(ElementTree.fromstring(content.decode()))
        schema_key = DEVICE
        root = ROOT
        for k in xml_dict.keys():
            m = XML_ROOT_SANITY_PATTERN.findall(k)
            if len(m) == 3 and m[1][0] and m[2][5]:
                schema_key = m[1][0]
                root = m[2][5]
                break
        return flatten_keys(xml_dict, "{%s}" % schema_key)[root]
    return {}



import logging
from typing import List

log = logging.getLogger(__name__)


class CaseInsensitive:
    def __init__(self, **kwargs) -> None:
        not_evaluated = {}
        for k, v in kwargs.items():
            if k.startswith("_"):
                not_evaluated[k] = v
                continue
            try:
                getattr(self, k)
                setattr(self, k, v)
            except AttributeError as err:
                not_evaluated[k] = v
        if not_evaluated:
            log.debug("%s did not apply kwargs: %s", self.__class__.__name__, not_evaluated)

    def _get_attr_name(self, case_insensitive: str) -> str:
        for k, v in self.__dict__.items():
            if k.lower() == case_insensitive.lower():
                return k
        raise AttributeError(case_insensitive)

    def __getattr__(self, item):
        if item in self.__dict__:
            return self.__dict__[item]
        for k, v in self.__class__.__dict__.items():
            if k.lower() == item.lower():
                if k not in self.__dict__:
                    self.__dict__[k] = v
                return v
        raise AttributeError(item)

    def __setattr__(self, item, value):
        if item in self.__dict__:
            self.__dict__[item] = value
            return
        to_update = None
        for k, v in self.__dict__.items():
            if k.lower() == item.lower():
                to_update = k
                break
        self.__dict__[to_update or item] = value

    def as_dict(self) -> dict:
        return {
            k: v for k, v in self.__dict__.items() if not k.startswith("_") and not callable(v)
        }


class Service(CaseInsensitive):
    serviceType = None
    serviceId = None
    controlURL = None
    eventSubURL = None
    SCPDURL = None


class Device(CaseInsensitive):
    serviceList = None
    deviceList = None
    deviceType = None
    friendlyName = None
    manufacturer = None
    manufacturerURL = None
    modelDescription = None
    modelName = None
    modelNumber = None
    modelURL = None
    serialNumber = None
    udn = None
    upc = None
    presentationURL = None
    iconList = None

    def __init__(self, devices: List, services: List, **kwargs) -> None:
        super(Device, self).__init__(**kwargs)
        if self.serviceList and "service" in self.serviceList:
            new_services = self.serviceList["service"]
            if isinstance(new_services, dict):
                new_services = [new_services]
            services.extend([Service(**service) for service in new_services])
        if self.deviceList:
            for kw in self.deviceList.values():
                if isinstance(kw, dict):
                    d = Device(devices, services, **kw)
                    devices.append(d)
                elif isinstance(kw, list):
                    for _inner_kw in kw:
                        d = Device(devices, services, **_inner_kw)
                        devices.append(d)
                else:
                    log.warning("failed to parse device:\n%s", kw)



from aioupnp.util import flatten_keys
from aioupnp.constants import FAULT, CONTROL


class UPnPError(Exception):
    pass


def handle_fault(response: dict) -> dict:
    if FAULT in response:
        fault = flatten_keys(response[FAULT], "{%s}" % CONTROL)
        error_description = fault['detail']['UPnPError']['errorDescription']
        raise UPnPError(error_description)
    return response



import logging
import socket
import asyncio
from collections import OrderedDict
from typing import Dict, List, Union, Type
from aioupnp.util import get_dict_val_case_insensitive, BASE_PORT_REGEX, BASE_ADDRESS_REGEX
from aioupnp.constants import SPEC_VERSION, SERVICE
from aioupnp.commands import SOAPCommands
from aioupnp.device import Device, Service
from aioupnp.protocols.ssdp import fuzzy_m_search, m_search
from aioupnp.protocols.scpd import scpd_get
from aioupnp.protocols.soap import SOAPCommand
from aioupnp.serialization.ssdp import SSDPDatagram
from aioupnp.util import flatten_keys
from aioupnp.fault import UPnPError

log = logging.getLogger(__name__)

return_type_lambas = {
    Union[None, str]: lambda x: x if x is not None and str(x).lower() not in ['none', 'nil'] else None
}


def get_action_list(element_dict: dict) -> List:  # [(<method>, [<input1>, ...], [<output1, ...]), ...]
    service_info = flatten_keys(element_dict, "{%s}" % SERVICE)
    if "actionList" in service_info:
        action_list = service_info["actionList"]
    else:
        return []
    if not len(action_list):  # it could be an empty string
        return []

    result: list = []
    if isinstance(action_list["action"], dict):
        arg_dicts = action_list["action"]['argumentList']['argument']
        if not isinstance(arg_dicts, list):  # when there is one arg
            arg_dicts = [arg_dicts]
        return [[
            action_list["action"]['name'],
            [i['name'] for i in arg_dicts if i['direction'] == 'in'],
            [i['name'] for i in arg_dicts if i['direction'] == 'out']
        ]]
    for action in action_list["action"]:
        if not action.get('argumentList'):
            result.append((action['name'], [], []))
        else:
            arg_dicts = action['argumentList']['argument']
            if not isinstance(arg_dicts, list):  # when there is one arg
                arg_dicts = [arg_dicts]
            result.append((
                action['name'],
                [i['name'] for i in arg_dicts if i['direction'] == 'in'],
                [i['name'] for i in arg_dicts if i['direction'] == 'out']
            ))
    return result


class Gateway:
    def __init__(self, ok_packet: SSDPDatagram, m_search_args: OrderedDict, lan_address: str,
                 gateway_address: str) -> None:
        self._ok_packet = ok_packet
        self._m_search_args = m_search_args
        self._lan_address = lan_address
        self.usn = (ok_packet.usn or '').encode()
        self.ext = (ok_packet.ext or '').encode()
        self.server = (ok_packet.server or '').encode()
        self.location = (ok_packet.location or '').encode()
        self.cache_control = (ok_packet.cache_control or '').encode()
        self.date = (ok_packet.date or '').encode()
        self.urn = (ok_packet.st or '').encode()

        self._xml_response = b""
        self._service_descriptors: Dict = {}
        self.base_address = BASE_ADDRESS_REGEX.findall(self.location)[0]
        self.port = int(BASE_PORT_REGEX.findall(self.location)[0])
        self.base_ip = self.base_address.lstrip(b"http://").split(b":")[0]
        assert self.base_ip == gateway_address.encode()
        self.path = self.location.split(b"%s:%i/" % (self.base_ip, self.port))[1]

        self.spec_version = None
        self.url_base = None

        self._device: Union[None, Device] = None
        self._devices: List = []
        self._services: List = []

        self._unsupported_actions: Dict = {}
        self._registered_commands: Dict = {}
        self.commands = SOAPCommands()

    def gateway_descriptor(self) -> dict:
        r = {
            'server': self.server.decode(),
            'urlBase': self.url_base,
            'location': self.location.decode(),
            "specVersion": self.spec_version,
            'usn': self.usn.decode(),
            'urn': self.urn.decode(),
        }
        return r

    @property
    def manufacturer_string(self) -> str:
        if not self.devices:
            raise NotImplementedError()
        device = list(self.devices.values())[0]
        return "%s %s" % (device.manufacturer, device.modelName)

    @property
    def services(self) -> Dict:
        if not self._device:
            return {}
        return {service.serviceType: service for service in self._services}

    @property
    def devices(self) -> Dict:
        if not self._device:
            return {}
        return {device.udn: device for device in self._devices}

    def get_service(self, service_type: str) -> Union[Type[Service], None]:
        for service in self._services:
            if service.serviceType.lower() == service_type.lower():
                return service
        return None

    @property
    def soap_requests(self) -> List:
        soap_call_infos = []
        for name in self._registered_commands.keys():
            if not hasattr(getattr(self.commands, name), "_requests"):
                continue
            soap_call_infos.extend([
                (name, request_args, raw_response, decoded_response, soap_error, ts)
                for (
                    request_args, raw_response, decoded_response, soap_error, ts
                ) in getattr(self.commands, name)._requests
            ])
        soap_call_infos.sort(key=lambda x: x[5])
        return soap_call_infos

    def debug_gateway(self) -> Dict:
        return {
            'manufacturer_string': self.manufacturer_string,
            'gateway_address': self.base_ip,
            'gateway_descriptor': self.gateway_descriptor(),
            'gateway_xml': self._xml_response,
            'services_xml': self._service_descriptors,
            'services': {service.SCPDURL: service.as_dict() for service in self._services},
            'm_search_args': [(k, v) for (k, v) in self._m_search_args.items()],
            'reply': self._ok_packet.as_dict(),
            'soap_port': self.port,
            'registered_soap_commands': self._registered_commands,
            'unsupported_soap_commands': self._unsupported_actions,
            'soap_requests': self.soap_requests
        }

    @classmethod
    async def _discover_gateway(cls, lan_address: str, gateway_address: str, timeout: int = 30,
                               igd_args: OrderedDict = None,  ssdp_socket: socket.socket = None,
                               soap_socket: socket.socket = None, unicast: bool = False):
        ignored: set = set()
        while True:
            if not igd_args:
                m_search_args, datagram = await asyncio.wait_for(fuzzy_m_search(lan_address, gateway_address, timeout, ssdp_socket,
                                                               ignored, unicast), timeout)
            else:
                m_search_args = OrderedDict(igd_args)
                datagram = await m_search(lan_address, gateway_address, igd_args, timeout, ssdp_socket, ignored,
                                          unicast)
            try:
                gateway = cls(datagram, m_search_args, lan_address, gateway_address)
                await gateway.discover_commands(soap_socket)
                log.debug('found gateway device %s', datagram.location)
                return gateway
            except (asyncio.TimeoutError, UPnPError):
                log.debug("get %s timed out, looking for other devices", datagram.location)
                ignored.add(datagram.location)
                continue

    @classmethod
    async def discover_gateway(cls, lan_address: str, gateway_address: str, timeout: int = 30,
                               igd_args: OrderedDict = None,  ssdp_socket: socket.socket = None,
                               soap_socket: socket.socket = None, unicast: bool = None):
        if unicast is not None:
            return await cls._discover_gateway(lan_address, gateway_address, timeout, igd_args, ssdp_socket,
                                               soap_socket, unicast=unicast)
        done, pending = await asyncio.wait([
            cls._discover_gateway(
                lan_address, gateway_address, timeout, igd_args, ssdp_socket, soap_socket, unicast=True
            ),
            cls._discover_gateway(
                lan_address, gateway_address, timeout, igd_args, ssdp_socket, soap_socket, unicast=False
            )], return_when=asyncio.tasks.FIRST_COMPLETED
        )
        for task in list(pending):
            task.cancel()
        result = list(done)[0].result()
        return result

    async def discover_commands(self, soap_socket: socket.socket = None):
        response, xml_bytes, get_err = await scpd_get(self.path.decode(), self.base_ip.decode(), self.port)
        self._xml_response = xml_bytes
        if get_err is not None:
            raise get_err
        self.spec_version = get_dict_val_case_insensitive(response, SPEC_VERSION)
        self.url_base = get_dict_val_case_insensitive(response, "urlbase")
        if not self.url_base:
            self.url_base = self.base_address.decode()
        if response:
            self._device = Device(
                self._devices, self._services, **get_dict_val_case_insensitive(response, "device")
            )
        else:
            self._device = Device(self._devices, self._services)
        for service_type in self.services.keys():
            await self.register_commands(self.services[service_type], soap_socket)

    async def register_commands(self, service: Service, soap_socket: socket.socket = None):
        if not service.SCPDURL:
            raise UPnPError("no scpd url")
        service_dict, xml_bytes, get_err = await scpd_get(service.SCPDURL, self.base_ip.decode(), self.port)
        self._service_descriptors[service.SCPDURL] = xml_bytes

        if get_err is not None:
            raise get_err
        if not service_dict:
            return

        action_list = get_action_list(service_dict)

        for name, inputs, outputs in action_list:
            try:
                current = getattr(self.commands, name)
                annotations = current.__annotations__
                return_types = annotations.get('return', None)
                if return_types:
                    if hasattr(return_types, '__args__'):
                        return_types = tuple([return_type_lambas.get(a, a) for a in return_types.__args__])
                    elif isinstance(return_types, type):
                        return_types = (return_types, )
                    return_types = {r: t for r, t in zip(outputs, return_types)}
                param_types = {}
                for param_name, param_type in annotations.items():
                    if param_name == "return":
                        continue
                    param_types[param_name] = param_type
                command = SOAPCommand(
                    self.base_ip.decode(), self.port, service.controlURL, service.serviceType.encode(),
                    name, param_types, return_types, inputs, outputs, soap_socket)
                setattr(command, "__doc__", current.__doc__)
                setattr(self.commands, command.method, command)

                self._registered_commands[command.method] = service.serviceType
                log.debug("registered %s::%s", service.serviceType, command.method)
            except AttributeError:
                s = self._unsupported_actions.get(service.serviceType, [])
                s.append(name)
                self._unsupported_actions[service.serviceType] = s
                log.debug("available command for %s does not have a wrapper implemented: %s %s %s",
                          service.serviceType, name, inputs, outputs)



import unittest
from aioupnp.serialization.scpd import serialize_scpd_get


class TestSCPDSerialization(unittest.TestCase):
    path, lan_address = '/InternetGatewayDevice.xml', '10.0.0.1'
    expected_result = b'GET /InternetGatewayDevice.xml HTTP/1.1\r\n' \
                      b'Accept-Encoding: gzip\r\nHost: 10.0.0.1\r\nConnection: Close\r\n\r\n'

    def test_serialize_get(self):
        self.assertEqual(serialize_scpd_get(self.path, self.lan_address), self.expected_result)



import re
from xml.etree import ElementTree
from aioupnp.util import etree_to_dict, flatten_keys
from aioupnp.fault import handle_fault, UPnPError
from aioupnp.constants import XML_VERSION, ENVELOPE, BODY

CONTENT_NO_XML_VERSION_PATTERN = re.compile(
    "(\<s\:Envelope xmlns\:s=\"http\:\/\/schemas\.xmlsoap\.org\/soap\/envelope\/\"(\s*.)*\>)".encode()
)


def serialize_soap_post(method: str, param_names: list, service_id: bytes, gateway_address: bytes,
                        control_url: bytes, **kwargs) -> bytes:
    args = "".join("<%s>%s</%s>" % (n, kwargs.get(n), n) for n in param_names)
    soap_body = ('\r\n%s\r\n<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" '
                 's:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/"><s:Body>'
                 '<u:%s xmlns:u="%s">%s</u:%s></s:Body></s:Envelope>' % (
                     XML_VERSION, method, service_id.decode(),
                     args, method))
    if "http://" in gateway_address.decode():
        host = gateway_address.decode().split("http://")[1]
    else:
        host = gateway_address.decode()
    return (
            (
                'POST %s HTTP/1.1\r\n'
                'Host: %s\r\n'
                'User-Agent: python3/aioupnp, UPnP/1.0, MiniUPnPc/1.9\r\n'
                'Content-Length: %i\r\n'
                'Content-Type: text/xml\r\n'
                'SOAPAction: \"%s#%s\"\r\n'
                'Connection: Close\r\n'
                'Cache-Control: no-cache\r\n'
                'Pragma: no-cache\r\n'
                '%s'
                '\r\n'
            ) % (
                control_url.decode(),  # could be just / even if it shouldn't be
                host,
                len(soap_body),
                service_id.decode(),  # maybe no quotes
                method,
                soap_body
            )
    ).encode()


def deserialize_soap_post_response(response: bytes, method: str, service_id: str) -> dict:
    parsed = CONTENT_NO_XML_VERSION_PATTERN.findall(response)
    content = b'' if not parsed else parsed[0][0]
    content_dict = etree_to_dict(ElementTree.fromstring(content.decode()))
    envelope = content_dict[ENVELOPE]
    response_body = flatten_keys(envelope[BODY], "{%s}" % service_id)
    body = handle_fault(response_body)  # raises UPnPError if there is a fault
    response_key = None
    if not body:
        return {}
    for key in body:
        if method in key:
            response_key = key
            break
    if not response_key:
        raise UPnPError("unknown response fields for %s: %s" % (method, body))
    return body[response_key]



import unittest
from aioupnp.serialization.soap import serialize_soap_post


class TestSOAPSerialization(unittest.TestCase):
    param_names: list = []
    kwargs: dict = {}
    method, gateway_address = "GetExternalIPAddress", b'10.0.0.1'
    st, lan_address, path = b'urn:schemas-upnp-org:service:WANIPConnection:1', '10.0.0.1', b'/soap.cgi?service=WANIPConn1'
    expected_result = b'POST /soap.cgi?service=WANIPConn1 HTTP/1.1\r\n' \
                      b'Host: 10.0.0.1\r\nUser-Agent: python3/aioupnp, UPnP/1.0, MiniUPnPc/1.9\r\n' \
                      b'Content-Length: 285\r\nContent-Type: text/xml\r\n' \
                      b'SOAPAction: "urn:schemas-upnp-org:service:WANIPConnection:1#GetExternalIPAddress"\r\n' \
                      b'Connection: Close\r\nCache-Control: no-cache\r\nPragma: no-cache\r\n\r\n' \
                      b'<?xml version="1.0"?>\r\n<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"' \
                      b' s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">' \
                      b'<s:Body><u:GetExternalIPAddress xmlns:u="urn:schemas-upnp-org:service:WANIPConnection:1">' \
                      b'</u:GetExternalIPAddress></s:Body></s:Envelope>\r\n'

    def test_serialize_get(self):
        self.assertEqual(serialize_soap_post(
            self.method, self.param_names, self.st, self.gateway_address, self.path, **self.kwargs
        ), self.expected_result)



import os
from setuptools import setup, find_packages  # type: ignore
from aioupnp import __version__, __name__, __email__, __author__, __license__

console_scripts = [
    'aioupnp = aioupnp.__main__:main',
]

package_name = "aioupnp"
base_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(base_dir, 'README.md'), 'rb') as f:
    long_description = f.read().decode('utf-8')

setup(
    name=__name__,
    version=__version__,
    author=__author__,
    author_email=__email__,
    description="UPnP for asyncio",
    keywords="upnp asyncio",
    long_description=long_description,
    url="https://github.com/lbryio/aioupnp",
    license=__license__,
    packages=find_packages(),
    entry_points={'console_scripts': console_scripts},
    install_requires=[
        'netifaces',
    ],
)



import logging
import socket
import asyncio
import typing
import time
from aioupnp.protocols.scpd import scpd_post
from aioupnp.fault import UPnPError

log = logging.getLogger(__name__)


def safe_type(t):
    if t is typing.Tuple:
        return tuple
    if t is typing.List:
        return list
    if t is typing.Dict:
        return dict
    if t is typing.Set:
        return set
    return t


class SOAPCommand:
    def __init__(self, gateway_address: str, service_port: int, control_url: str, service_id: bytes, method: str,
                 param_types: dict, return_types: dict, param_order: list, return_order: list,
                 soap_socket: socket.socket = None) -> None:
        self.gateway_address = gateway_address
        self.service_port = service_port
        self.control_url = control_url
        self.service_id = service_id
        self.method = method
        self.param_types = param_types
        self.param_order = param_order
        self.return_types = return_types
        self.return_order = return_order
        self.soap_socket = soap_socket
        self._requests: typing.List = []

    async def __call__(self, **kwargs) -> typing.Union[None, typing.Dict, typing.List, typing.Tuple]:
        if set(kwargs.keys()) != set(self.param_types.keys()):
            raise Exception("argument mismatch: %s vs %s" % (kwargs.keys(), self.param_types.keys()))
        soap_kwargs = {n: safe_type(self.param_types[n])(kwargs[n]) for n in self.param_types.keys()}
        response, xml_bytes, err = await scpd_post(
            self.control_url, self.gateway_address, self.service_port, self.method, self.param_order,
            self.service_id, self.soap_socket, **soap_kwargs
        )
        if err is not None:
            self._requests.append((soap_kwargs, xml_bytes, None, err, time.time()))
            raise err
        if not response:
            result = None
        else:
            recast_result = tuple([safe_type(self.return_types[n])(response.get(n)) for n in self.return_order])
            if len(recast_result) == 1:
                result = recast_result[0]
            else:
                result = recast_result
        self._requests.append((soap_kwargs, xml_bytes, result, None, time.time()))
        return result



import os
import socket
import logging
import json
import asyncio
import zlib
import base64
from collections import OrderedDict
from typing import Tuple, Dict, List, Union
from aioupnp.fault import UPnPError
from aioupnp.gateway import Gateway
from aioupnp.util import get_gateway_and_lan_addresses
from aioupnp.protocols.ssdp import m_search, fuzzy_m_search
from aioupnp.protocols.soap import SOAPCommand
from aioupnp.serialization.ssdp import SSDPDatagram

log = logging.getLogger(__name__)


def cli(fn):
    fn._cli = True
    return fn


def _encode(x):
    if isinstance(x, bytes):
        return x.decode()
    elif isinstance(x, Exception):
        return str(x)
    return x


class UPnP:
    def __init__(self, lan_address: str, gateway_address: str, gateway: Gateway) -> None:
        self.lan_address = lan_address
        self.gateway_address = gateway_address
        self.gateway = gateway

    @classmethod
    def get_lan_and_gateway(cls, lan_address: str = '', gateway_address: str = '',
                            interface_name: str = 'default') -> Tuple[str, str]:
        if not lan_address or not gateway_address:
            gateway_addr, lan_addr = get_gateway_and_lan_addresses(interface_name)
            lan_address = lan_address or lan_addr
            gateway_address = gateway_address or gateway_addr
        return lan_address, gateway_address

    @classmethod
    async def discover(cls, lan_address: str = '', gateway_address: str = '', timeout: int = 30,
                       igd_args: OrderedDict = None, interface_name: str = 'default',
                       ssdp_socket: socket.socket = None, soap_socket: socket.socket = None):
        try:
            lan_address, gateway_address = cls.get_lan_and_gateway(lan_address, gateway_address, interface_name)
        except Exception as err:
            raise UPnPError("failed to get lan and gateway addresses: %s" % str(err))
        gateway = await Gateway.discover_gateway(
            lan_address, gateway_address, timeout, igd_args, ssdp_socket, soap_socket
        )
        return cls(lan_address, gateway_address, gateway)

    @classmethod
    @cli
    async def m_search(cls, lan_address: str = '', gateway_address: str = '', timeout: int = 1,
                       igd_args: OrderedDict = None, interface_name: str = 'default',
                       ssdp_socket: socket.socket = None) -> Dict:
        try:
            lan_address, gateway_address = cls.get_lan_and_gateway(lan_address, gateway_address, interface_name)
            assert gateway_address and lan_address
        except Exception as err:
            raise UPnPError("failed to get lan and gateway addresses for interface \"%s\": %s" % (interface_name,
                                                                                                  str(err)))
        if not igd_args:
            igd_args, datagram = await fuzzy_m_search(lan_address, gateway_address, timeout, ssdp_socket)
        else:
            igd_args = OrderedDict(igd_args)
            datagram = await m_search(lan_address, gateway_address, igd_args, timeout, ssdp_socket)
        return {
            'lan_address': lan_address,
            'gateway_address': gateway_address,
            'm_search_kwargs': SSDPDatagram("M-SEARCH", igd_args).get_cli_igd_kwargs(),
            'discover_reply': datagram.as_dict()
        }

    @cli
    async def get_external_ip(self) -> str:
        return await self.gateway.commands.GetExternalIPAddress()

    @cli
    async def add_port_mapping(self, external_port: int, protocol: str, internal_port: int, lan_address: str,
                               description: str) -> None:
        return await self.gateway.commands.AddPortMapping(
            NewRemoteHost='', NewExternalPort=external_port, NewProtocol=protocol,
            NewInternalPort=internal_port, NewInternalClient=lan_address,
            NewEnabled=1, NewPortMappingDescription=description, NewLeaseDuration='0'
        )

    @cli
    async def get_port_mapping_by_index(self, index: int) -> Dict:
        result = await self._get_port_mapping_by_index(index)
        if result:
            if isinstance(self.gateway.commands.GetGenericPortMappingEntry, SOAPCommand):
                return {
                    k: v for k, v in zip(self.gateway.commands.GetGenericPortMappingEntry.return_order, result)
                }
        return {}

    async def _get_port_mapping_by_index(self, index: int) -> Union[None, Tuple[Union[None, str], int, str,
                                                                                int, str, bool, str, int]]:
        try:
            redirect = await self.gateway.commands.GetGenericPortMappingEntry(NewPortMappingIndex=index)
            return redirect
        except UPnPError:
            return None

    @cli
    async def get_redirects(self) -> List[Dict]:
        redirects = []
        cnt = 0
        redirect = await self.get_port_mapping_by_index(cnt)
        while redirect:
            redirects.append(redirect)
            cnt += 1
            redirect = await self.get_port_mapping_by_index(cnt)
        return redirects

    @cli
    async def get_specific_port_mapping(self, external_port: int, protocol: str) -> Dict:
        """
        :param external_port: (int) external port to listen on
        :param protocol:      (str) 'UDP' | 'TCP'
        :return: (int) <internal port>, (str) <lan ip>, (bool) <enabled>, (str) <description>, (int) <lease time>
        """

        try:
            result = await self.gateway.commands.GetSpecificPortMappingEntry(
                NewRemoteHost='', NewExternalPort=external_port, NewProtocol=protocol
            )
            if result and isinstance(self.gateway.commands.GetSpecificPortMappingEntry, SOAPCommand):
                return {k: v for k, v in zip(self.gateway.commands.GetSpecificPortMappingEntry.return_order, result)}
        except UPnPError:
            pass
        return {}

    @cli
    async def delete_port_mapping(self, external_port: int, protocol: str) -> None:
        """
        :param external_port: (int) external port to listen on
        :param protocol:      (str) 'UDP' | 'TCP'
        :return: None
        """
        return await self.gateway.commands.DeletePortMapping(
            NewRemoteHost="", NewExternalPort=external_port, NewProtocol=protocol
        )

    @cli
    async def get_next_mapping(self, port: int, protocol: str, description: str, internal_port: int=None) -> int:
        if protocol not in ["UDP", "TCP"]:
            raise UPnPError("unsupported protocol: {}".format(protocol))
        internal_port = internal_port or port
        redirect_tups = []
        cnt = 0
        port = int(port)
        internal_port = int(internal_port)
        redirect = await self._get_port_mapping_by_index(cnt)
        while redirect:
            redirect_tups.append(redirect)
            cnt += 1
            redirect = await self._get_port_mapping_by_index(cnt)

        redirects = {
            "%i:%s" % (ext_port, proto): (int_host, int_port, desc)
            for (ext_host, ext_port, proto, int_port, int_host, enabled, desc, _) in redirect_tups
        }
        while ("%i:%s" % (port, protocol)) in redirects:
            int_host, int_port, _ = redirects["%i:%s" % (port, protocol)]
            if int_host == self.lan_address and int_port == internal_port:
                break
            port += 1

        await self.add_port_mapping(  # set one up
                port, protocol, internal_port, self.lan_address, description
        )
        return port

    @cli
    async def debug_gateway(self) -> str:
        return json.dumps({
            "gateway": self.gateway.debug_gateway(),
            "client_address": self.lan_address,
        }, default=_encode, indent=2)

    @property
    def zipped_debugging_info(self) -> str:
        return base64.b64encode(zlib.compress(
            json.dumps({
                "gateway": self.gateway.debug_gateway(),
                "client_address": self.lan_address,
            }, default=_encode, indent=2).encode()
        )).decode()

    @cli
    async def generate_test_data(self):
        print("found gateway via M-SEARCH")
        try:
            external_ip = await self.get_external_ip()
            print("got external ip: %s" % external_ip)
        except (UPnPError, NotImplementedError):
            print("failed to get the external ip")
        try:
            await self.get_redirects()
            print("got redirects")
        except (UPnPError, NotImplementedError):
            print("failed to get redirects")
        try:
            await self.get_specific_port_mapping(4567, "UDP")
            print("got specific mapping")
        except (UPnPError, NotImplementedError):
            print("failed to get specific mapping")
        try:
            ext_port = await self.get_next_mapping(4567, "UDP", "aioupnp test mapping")
            print("set up external mapping to port %i" % ext_port)
            try:
                await self.get_specific_port_mapping(4567, "UDP")
                print("got specific mapping")
            except (UPnPError, NotImplementedError):
                print("failed to get specific mapping")
            try:
                await self.get_redirects()
                print("got redirects")
            except (UPnPError, NotImplementedError):
                print("failed to get redirects")
            await self.delete_port_mapping(ext_port, "UDP")
            print("deleted mapping")
        except (UPnPError, NotImplementedError):
            print("failed to add and remove a mapping")
        try:
            await self.get_redirects()
            print("got redirects")
        except (UPnPError, NotImplementedError):
            print("failed to get redirects")
        try:
            await self.get_specific_port_mapping(4567, "UDP")
            print("got specific mapping")
        except (UPnPError, NotImplementedError):
            print("failed to get specific mapping")
        if self.gateway.devices:
            device = list(self.gateway.devices.values())[0]
            assert device.manufacturer and device.modelName
            device_path = os.path.join(os.getcwd(), self.gateway.manufacturer_string)
        else:
            device_path = os.path.join(os.getcwd(), "UNKNOWN GATEWAY")
        with open(device_path, "w") as f:
            f.write(await self.debug_gateway())
        return "Generated test data! -> %s" % device_path

    @cli
    async def get_natrsip_status(self) -> Tuple[bool, bool]:
        """Returns (NewRSIPAvailable, NewNATEnabled)"""
        return await self.gateway.commands.GetNATRSIPStatus()

    @cli
    async def set_connection_type(self, NewConnectionType: str) -> None:
        """Returns None"""
        return await self.gateway.commands.SetConnectionType(NewConnectionType)

    @cli
    async def get_connection_type_info(self) -> Tuple[str, str]:
        """Returns (NewConnectionType, NewPossibleConnectionTypes)"""
        return await self.gateway.commands.GetConnectionTypeInfo()

    @cli
    async def get_status_info(self) -> Tuple[str, str, int]:
        """Returns (NewConnectionStatus, NewLastConnectionError, NewUptime)"""
        return await self.gateway.commands.GetStatusInfo()

    @cli
    async def force_termination(self) -> None:
        """Returns None"""
        return await self.gateway.commands.ForceTermination()

    @cli
    async def request_connection(self) -> None:
        """Returns None"""
        return await self.gateway.commands.RequestConnection()

    @cli
    async def get_common_link_properties(self):
        """Returns (NewWANAccessType, NewLayer1UpstreamMaxBitRate, NewLayer1DownstreamMaxBitRate, NewPhysicalLinkStatus)"""
        return await self.gateway.commands.GetCommonLinkProperties()

    @cli
    async def get_total_bytes_sent(self):
        """Returns (NewTotalBytesSent)"""
        return await self.gateway.commands.GetTotalBytesSent()

    @cli
    async def get_total_bytes_received(self):
        """Returns (NewTotalBytesReceived)"""
        return await self.gateway.commands.GetTotalBytesReceived()

    @cli
    async def get_total_packets_sent(self):
        """Returns (NewTotalPacketsSent)"""
        return await self.gateway.commands.GetTotalPacketsSent()

    @cli
    async def get_total_packets_received(self):
        """Returns (NewTotalPacketsReceived)"""
        return await self.gateway.commands.GetTotalPacketsReceived()

    @cli
    async def x_get_ics_statistics(self) -> Tuple[int, int, int, int, str, str]:
        """Returns (TotalBytesSent, TotalBytesReceived, TotalPacketsSent, TotalPacketsReceived, Layer1DownstreamMaxBitRate, Uptime)"""
        return await self.gateway.commands.X_GetICSStatistics()

    @cli
    async def get_default_connection_service(self):
        """Returns (NewDefaultConnectionService)"""
        return await self.gateway.commands.GetDefaultConnectionService()

    @cli
    async def set_default_connection_service(self, NewDefaultConnectionService: str) -> None:
        """Returns (None)"""
        return await self.gateway.commands.SetDefaultConnectionService(NewDefaultConnectionService)

    @cli
    async def set_enabled_for_internet(self, NewEnabledForInternet: bool) -> None:
        return await self.gateway.commands.SetEnabledForInternet(NewEnabledForInternet)

    @cli
    async def get_enabled_for_internet(self) -> bool:
        return await self.gateway.commands.GetEnabledForInternet()

    @cli
    async def get_maximum_active_connections(self, NewActiveConnectionIndex: int):
        return await self.gateway.commands.GetMaximumActiveConnections(NewActiveConnectionIndex)

    @cli
    async def get_active_connections(self) -> Tuple[str, str]:
        """Returns (NewActiveConnDeviceContainer, NewActiveConnectionServiceID"""
        return await self.gateway.commands.GetActiveConnections()

    @classmethod
    def run_cli(cls, method, igd_args: OrderedDict, lan_address: str = '', gateway_address: str = '', timeout: int = 30,
                interface_name: str = 'default', kwargs: dict = None) -> None:
        kwargs = kwargs or {}
        igd_args = igd_args
        timeout = int(timeout)
        close_loop = False
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            close_loop = True
        if not loop and not close_loop:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            close_loop = True

        fut: asyncio.Future = asyncio.Future()

        async def wrapper():
            if method == 'm_search':
                fn = lambda *_a, **_kw: cls.m_search(
                    lan_address, gateway_address, timeout, igd_args, interface_name
                )
            else:
                try:
                    u = await cls.discover(
                        lan_address, gateway_address, timeout, igd_args, interface_name
                    )
                except UPnPError as err:
                    fut.set_exception(err)
                    return
                if hasattr(u, method) and hasattr(getattr(u, method), "_cli"):
                    fn = getattr(u, method)
                else:
                    fut.set_exception(UPnPError("\"%s\" is not a recognized command" % method))
                    return
            try:
                result = await fn(**{k: fn.__annotations__[k](v) for k, v in kwargs.items()})
                fut.set_result(result)
            except UPnPError as err:
                fut.set_exception(err)

            except Exception as err:
                log.exception("uncaught error")
                fut.set_exception(UPnPError("uncaught error: %s" % str(err)))

        if not hasattr(UPnP, method) or not hasattr(getattr(UPnP, method), "_cli"):
            fut.set_exception(UPnPError("\"%s\" is not a recognized command" % method))
            wrapper = lambda : None

        loop.run_until_complete(wrapper())
        if close_loop:
            loop.close()
        try:
            result = fut.result()
        except UPnPError as err:
            print("aioupnp encountered an error:\n%s" % str(err))
            return

        if isinstance(result, (list, tuple, dict)):
            print(json.dumps(result, indent=2, default=_encode))
        else:
            print(result)
